// threads.c

/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

#include "../common/cmdlib.h"	// gb
#include "threads.h"
#include <sys/types.h>
#include <sys/sysctl.h>

#define MAX_THREADS	64

int getProcessorCount()
{
	int count;
	size_t size = sizeof( count );
	
	if( sysctlbyname( "hw.ncpu", &count, &size, NULL, 0 ) )
	{
		return 1;
	}
	
	if( count > MAX_THREADS )
	{
		count = MAX_THREADS;
	}
	
	return count;
}

int		numthreads = 1;

#ifdef __alpha
pthread_mutex_t	*my_mutex;
#endif

void InitThreads()
{
#ifdef __alpha
	numthreads = (getProcessorCount() * 2);
	printf( "Using %d threads\n\n", numthreads );
	
	pthread_mutexattr_t	mattrib;
	
	my_mutex = (pthread_mutex_t*)malloc( sizeof(pthread_mutex_t) );
	
	if (pthread_mutexattr_init( &mattrib ) == -1)
	{
		Error ("pthread_mutexattr_init failed");
	}
	
	if (pthread_mutexattr_settype( &mattrib, PTHREAD_MUTEX_NORMAL ) == -1)
	{
		Error ("pthread_mutexattr_settype failed");
	}
	
	if (pthread_mutex_init( my_mutex, &mattrib ) == -1)
	{
		Error ("pthread_mutex_init failed");
	}
#endif
}

void* LightThread( void *junk );

/*
===============
RunThreadsOn
===============
*/
void RunThreadsOn()
{
#ifdef __alpha
	pthread_t	work_threads[MAX_THREADS];
	pthread_attr_t	attrib;
	int		i;
	
	if (numthreads == 1)
	{
		LightThread( NULL );
		return;
	}
	
	if( pthread_attr_init( &attrib ) == -1 )
	{
		Error ("pthread_attr_init failed");
	}
	
	if (pthread_attr_setstacksize (&attrib, 0x100000) == -1)
	{
		Error ("pthread_attr_setstacksize failed");
	}
	
	for (i=0 ; i<numthreads ; i++)
	{
		if (pthread_create(&work_threads[i], &attrib, LightThread, NULL) == -1)
		{
			Error ("pthread_create failed");
		}
	}
	
	for (i=0 ; i<numthreads ; i++)
	{
		if (pthread_join (work_threads[i], NULL) == -1)
		{
			Error ("pthread_join failed");
		}
	}
	
	// Destroy the mutex
	
	if (pthread_mutex_destroy( my_mutex ) == -1)
	{
		Error ("pthread_mutex_destroy failed");
	}
	
#else
	LightThread( NULL );
#endif
}
