/*

WAD source file

*/

// this is not id's - hence no GPL yet. gb

#include "qbsp.h"
#include <sys/types.h>
#include <dirent.h>
#include "wad.H"	// gb

// #include <io.h>	gb, sadly this stuff is Windows only
//			_findfirst(), _findclose() etc are OS specific.
//			on POSIX we should probably use fopen() etc.
//			There is also opendir() and friends.

/*
==================
WAD
==================
*/
WAD::WAD(void)
{
	cWads = 0;
	wadlist = NULL;
}

/*
==================
~WAD
==================
*/
WAD::~WAD(void)
{
	if (wadlist)
	{
		for (iWad=0; iWad<cWads; iWad++)
		{
			wadlist[iWad].Wad.Close();
			FreeMem(wadlist[iWad].lumps, OTHER, sizeof(lumpinfo_t) * wadlist[iWad].header.numlumps);
		}
		FreeMem(wadlist, OTHER, cWads * sizeof(wadlist_t));
	}
}

#define MAX_WADPATHSIZE 0x40000 // 256K

/*
==================
NeedArrayBytes
==================
*/

void WAD::NeedArrayBytes (int Size)
{
	if( Size >= MAX_WADPATHSIZE )
	{
		Message (msgError, errWadPathSize, MAX_WADPATHSIZE / 1024);
	}
}

/*
==================
ExpandWadPath
==================
*/

extern char		qdir[1024];
extern char		gamedir[1024];

char *WAD::ExpandWadPath (char *Path)
{
	//printf( "WAD : %s\n", Path );
	//printf( "MAP : %s\n", options.szMapName );
	//printf( "QUAKE DIR : %s\n", qdir );
	//printf( "GAME DIR : %s\n", gamedir );
	
	return Path;
}

/*
char *WAD::ExpandWadPath (char *Path)
{
	struct _finddata_t FindInfo;
	long		   Handle;
	int		   i, j, Result, Lth, Lth2;
	//char		   Src[1024], FullPath[1024], Drv[_MAX_DRIVE], Dir[_MAX_DIR];
	char		   Src[1024], FullPath[1024], Drv[3], Dir[256];
	char		   *WadPath, *PathBegin;
	bool		   WildCard = false;

	if (strchr (Path, '*') == NULL)
		return Path; // No wildcards
	
	WadPath = (char *)AllocMem (OTHER, MAX_WADPATHSIZE);

	Lth = strlen (Path);
	PathBegin = NULL;

	for (i = j = 0; i < Lth + 1; ++i)
	{
		if (PathBegin == NULL)
			PathBegin = &Path[i]; // Beginning of new wad path
	    
		NeedArrayBytes (j);
		WadPath[j++] = Path[i];
		
		if (Path[i] == '*')
			WildCard = true;

		if (Path[i] != ';' && Path[i] != '\0')
			continue;

		// New wad path
		if (!WildCard)
		{
			PathBegin = NULL;
			continue; // No wildcard in this path
		}

		// New wildcard wad path
		WildCard = false;
		Lth2 = &Path[i] - PathBegin;
		memcpy (Src, PathBegin, Lth2);
		Src[Lth2] = '\0';
		PathBegin = NULL;

		j -= Lth2 + 1;
		WadPath[j] = '\0';
		
		_splitpath (Src, Drv, Dir, NULL, NULL);

		Handle = _findfirst (Src, &FindInfo);

		if (Handle == -1L)
			continue; // No file match

		Result = 0;

		while (Result == 0)
		{
			_makepath (FullPath, Drv, Dir, FindInfo.name, NULL);

			Lth2 = strlen (FullPath);
		
			// Add filename to list
			NeedArrayBytes (j + Lth2 + 1);
			memcpy (&WadPath[j], FullPath, Lth2);
			j += Lth2;
			WadPath[j++] = ';';
			WadPath[j] = '\0';

			Result = _findnext (Handle, &FindInfo);
		}

		_findclose (Handle);
	}

	// Shrink WadPath to actual size
	WadPath = (char *)ReAllocOther (WadPath, strlen (WadPath) + 1);

	return WadPath;
}
*/

/*
==================
InitWADList
==================
*/
bool WAD::InitWADList(char *szWadList)
{
	wadlist_t *CWad;
	int	  i, len, WadLth;
	File	  *fileT;
	bool	  OpenOK, WadSizeOk, Wad2;
	char	  WadName[512], *szSaveList;

	if (!szWadList)
		return false;

	// Should never happen, but just in case...
	assert(!wadlist);

	// Expand wildcards in wad path, return path might be reallocated
	szWadList = ExpandWadPath (szSaveList = szWadList);

	WadLth = len = strlen (szWadList); // Save length for later

	// Count # of wads
	cWads = 1;
	for (i=0; i<len; i++)
		if (szWadList[i] == ';')
			cWads++;

	wadlist = (wadlist_t *)AllocMem (OTHER, cWads * sizeof(wadlist_t));

	if (len == 0)
		len = 1; // Special fix when no wad file specified

	// Verify that at least one WAD file exists
	i = iWad = 0;
	while (i < len)
	{
		szWadName = szWadList+i;
		while (szWadList[i] != 0 && szWadList[i] != ';')
			i++;
		szWadList[i] = 0;
		i++;

		CWad = &wadlist[iWad];
		fileT = &CWad->Wad;

		if (strlen(szWadName) == 0)
			OpenOK = false;
		else
			OpenOK = fileT->fOpen(szWadName, "rb", false);

		if (!OpenOK && iWad == 0 && i >= len)
		{
			// Check default WAD name
			strcpy(WadName, options.szMapName);
			StripExtension(WadName);
			DefaultExtension(WadName, ".wad");

			szWadName = WadName;

			OpenOK = fileT->fOpen(szWadName, "rb", false);

			if (OpenOK)
				Message(msgStat2, "Using default WAD: %s", szWadName);
		}

		if (OpenOK)
		{
			WadSizeOk = fileT->Length() >= sizeof(wadinfo_t);

			if (WadSizeOk)
			{
				fileT->Read(&CWad->header, sizeof(wadinfo_t));

				Wad2 = !strncmp(CWad->header.identification, "WAD2", 4);
				CWad->Wad3 = !strncmp(CWad->header.identification, "WAD3", 4);
			}

			if (!WadSizeOk || !Wad2 && !CWad->Wad3)
			{
				Message(msgWarning, warnNotWad, szWadName);
				fileT->Close();
			}
			else
			{
//				strcpy(CWad->szName, szWadName);
				fileT->Seek(CWad->header.infotableofs, SEEK_SET);
				CWad->lumps = (lumpinfo_t *)AllocMem(OTHER, sizeof(lumpinfo_t) * CWad->header.numlumps);
				fileT->Read(CWad->lumps, CWad->header.numlumps * sizeof(lumpinfo_t));
				iWad++;
				// Note that the file is NOT closed here!
				// Also iWad is only incremented for valid files
			}
		}
	}

	if (szWadList != szSaveList)
		FreeMem (szWadList, OTHER, WadLth + 1);

	// Remove invalid wads from memory
	wadlist = (wadlist_t *)ReAllocOther (wadlist, iWad * sizeof(wadlist_t));
	cWads = iWad;

	if (cWads > 0)
	{
		// Report missing textures early on
		AddAnimatingTextures ();
		
		for (i = 0; i < cMiptex; ++i)
		{
			for (iWad = 0; iWad < cWads; iWad++)
			{
				if (LoadLump (rgszMiptex[i], NULL) != 0)
					break;
			}
				
			if (iWad == cWads)
				Message(msgWarning, warnTextureNotFound, rgszMiptex[i]);
		}
	
		iWad = cWads; // Restore iWad
	}

	return iWad > 0;
}


/*
==================
fProcessWAD
==================
*/
void WAD::fProcessWAD(void)
{
	int i, j;
	dmiptexlump_t *l;

	if (cWads < 1)
		return;

	// Count texture size.  Slow but saves memory.
	for (i=0; i<cMiptex; i++)
		for (iWad = 0; iWad < cWads; iWad++)
		{
			for (j=0 ; j<wadlist[iWad].header.numlumps ; j++)
				if (!strcasecmp(rgszMiptex[i], wadlist[iWad].lumps[j].name))
				{
					// Found it. Add in the size and skip to outer loop.
// For EvlG			printf("Texture: %9s in wad: %s\n", rgszMiptex[i], wadlist[iWad].szName);
					pWorldEnt->cTexdata += wadlist[iWad].lumps[j].disksize;
					iWad = cWads;
					break;
				}

			// If we found the texture already, break out to outer loop
			if (j < wadlist[iWad].header.numlumps)
				break;
		}

	// Add space for dataofs array
	pWorldEnt->cTexdata += sizeof(int) * (cMiptex+1);

	// Default texture data to store in worldmodel
	pWorldEnt->pTexdata = (byte *)AllocMem(BSPTEX, pWorldEnt->cTexdata);
	l = (dmiptexlump_t *)pWorldEnt->pTexdata;
	l->nummiptex = cMiptex;

	LoadTextures(l);

	// Shrink tex size if needed, NOTE: array is not reallocated here
	if (pWorldEnt->iTexdata < pWorldEnt->cTexdata)
		pWorldEnt->cTexdata = pWorldEnt->iTexdata;

	// Last pass, mark unfound textures as such
	for (i=0; i<cMiptex; i++)
	{
		if (l->dataofs[i] == 0)
			l->dataofs[i] = -1;
	}
}

/*
==================
LoadTextures
==================
*/
void WAD::LoadTextures(dmiptexlump_t *l)
{
	int  i, len;
	byte *data;

	data = (byte *)&l->dataofs[cMiptex];
	pWorldEnt->iTexdata = data - (byte *)l;

	for (iWad = 0; iWad < cWads; iWad++)
	{
		for (i=0 ; i<cMiptex ; i++)
		{
			// Texture already found in a previous WAD
			if (l->dataofs[i] != 0)
				continue;

			l->dataofs[i] = data - (byte *)l;
			len = LoadLump(rgszMiptex[i], data);
					
			pWorldEnt->iTexdata += len;

			if (pWorldEnt->iTexdata > pWorldEnt->cTexdata)
				Message(msgError, errLowTextureCount);

			// didn't find the texture
			if (!len)
				l->dataofs[i] = 0;
			data += len;
		}
	}
}

/*
==================
LoadLump
==================
*/
int WAD::LoadLump(char *szName, byte *pDest)
{
	miptex2_t *MipTex;
	wadlist_t *CWad;
	int	  i, PalOffset;

	CWad = &wadlist[iWad];

	for (i = 0; i < CWad->header.numlumps; i++)
	{
		if (!strcasecmp(szName, CWad->lumps[i].name))
		{
			if (pDest == NULL)
			{
				// Only checking for missing textures
				if (options.fAllverbose || options.fTexverbose)
					Message (msgStat2, "Texture %s found in %s", szName, CWad->Wad.GetFName());

				return 1;
			}

			CWad->Wad.Seek(CWad->lumps[i].filepos, SEEK_SET);
			CWad->Wad.Read(pDest, CWad->lumps[i].disksize);
			CleanupName ((char *)pDest, (char *)pDest);

			if (CWad->Wad3)
			{
				// Check if any palette and reduce texture size accordingly
				MipTex = (miptex2_t *)pDest;

				PalOffset = MipTex->offsets[MIPLEVELS - 1] + MipTex->width * MipTex->height / 64;

				if (PalOffset < CWad->lumps[i].disksize)
					CWad->lumps[i].disksize = CWad->lumps[i].size = PalOffset;
			}
			
			return CWad->lumps[i].disksize;
		}
	}

	return 0;
}


/*
==================
AddAnimatingTextures
==================
*/
void WAD::AddAnimatingTextures (void)
{
	int	base;
	int	i, j, k, l;
	char	name[32];

	base = cMiptex;

	for (i=0 ; i<base ; i++)
	{
		if (rgszMiptex[i][0] != '+')
			continue;
		strcpy (name, rgszMiptex[i]);

		for (j=0 ; j<20 ; j++)
		{
			if (j < 10)
				name[1] = '0'+j;
			else
				name[1] = 'A'+j-10;		// alternate animation


			// see if this name exists in the wadfiles
			for (l=0; l<cWads; l++)
				for (k=0 ; k<wadlist[l].header.numlumps; k++)
					if (!strcasecmp(name, wadlist[l].lumps[k].name))
					{
						FindMiptex (name);	// add to the miptex list
						break;
					}
		}
	}

	if (cMiptex > base)
		Message(msgStat, "%6i texture frames added", cMiptex - base);
}


